import os
import sys
import matplotlib

# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')
import pylab

import caffe
from caffe import layers as L, params as P

# Import the DataValidator, which can then be used to
# validate and load the path to the already downloaded dataset.
sys.path.insert(0, '../../utils')
from data_utils import DataValidator


# Prepares network specification
def lenet(lmdb, batch_size):
    # Caffe's version of LeNet: a series of linear and simple nonlinear transformations
    n = caffe.NetSpec()

    n.data, n.label = L.Data(batch_size=batch_size, backend=P.Data.LMDB, source=lmdb,
                             transform_param=dict(scale=1. / 255), ntop=2)

    n.conv1 = L.Convolution(n.data, kernel_size=5, num_output=20, weight_filler=dict(type='xavier'))
    n.pool1 = L.Pooling(n.conv1, kernel_size=2, stride=2, pool=P.Pooling.MAX)
    n.conv2 = L.Convolution(n.pool1, kernel_size=5, num_output=50, weight_filler=dict(type='xavier'))
    n.pool2 = L.Pooling(n.conv2, kernel_size=2, stride=2, pool=P.Pooling.MAX)
    n.fc1 = L.InnerProduct(n.pool2, num_output=500, weight_filler=dict(type='xavier'))
    n.relu1 = L.ReLU(n.fc1, in_place=True)
    n.score = L.InnerProduct(n.relu1, num_output=10, weight_filler=dict(type='xavier'))
    n.loss = L.SoftmaxWithLoss(n.score, n.label)

    return n.to_proto()


# Names of the directories containing the LMDB files for TRAIN and TEST phases
test_dir = 'mnist/caffe/mnist_test_lmdb'
train_dir = 'mnist/caffe/mnist_train_lmdb'

# Validated path to the data root
DataValidator.validated_data_dir(train_dir)
data_dir = DataValidator.validated_data_dir(test_dir)

# Write the prototxt for TRAIN phase
with open('lenet_auto_train.prototxt', 'w') as f:
    f.write(str(lenet(os.path.join(data_dir, train_dir), 64)))

# Write the prototxt for TEST phase
with open('lenet_auto_test.prototxt', 'w') as f:
    f.write(str(lenet(os.path.join(data_dir, test_dir), 100)))

# Use the GPU for training
caffe.set_device(0)
caffe.set_mode_gpu()

# Load the solver and create train and test nets
solver = None  # ignore this workaround for lmdb data (can't instantiate two solvers on the same data)
solver = caffe.SGDSolver('lenet_auto_solver.prototxt')

solver.net.forward()  # train net
solver.test_nets[0].forward()  # test net (there can be more than one)

niter = 200
test_interval = 25
# losses will also be stored in the log
train_loss = pylab.zeros(niter)
test_acc = pylab.zeros(int(pylab.ceil(niter / test_interval)))
output = pylab.zeros((niter, 8, 10))

# the main solver loop
for it in range(niter):
    solver.step(1)  # SGD by Caffe

    # store the train loss
    train_loss[it] = solver.net.blobs['loss'].data

    # store the output on the first test batch
    # (start the forward pass at conv1 to avoid loading new data)
    solver.test_nets[0].forward(start='conv1')
    output[it] = solver.test_nets[0].blobs['score'].data[:8]

    # run a full test every so often
    # (Caffe can also do this for us and write to a log, but we show here
    #  how to do it directly in Python, where more complicated things are easier.)
    if it % test_interval == 0:
        print('Iteration', it, 'testing...')
        correct = 0
        for test_it in range(100):
            solver.test_nets[0].forward()
            correct += sum(solver.test_nets[0].blobs['score'].data.argmax(1)
                           == solver.test_nets[0].blobs['label'].data)
        test_acc[it // test_interval] = correct / 1e4

# Plot the training curve
_, ax1 = pylab.subplots()
ax2 = ax1.twinx()
ax1.plot(pylab.arange(niter), train_loss)
ax2.plot(test_interval * pylab.arange(len(test_acc)), test_acc, 'r')
ax1.set_xlabel('iteration')
ax1.set_ylabel('train loss')
ax2.set_ylabel('test accuracy')
ax2.set_title('Test Accuracy: {:.2f}'.format(test_acc[-1]))

# Save the plot to file. Use "bbox_inches='tight'" to remove surrounding whitespace
pylab.savefig('learning_curve.png', bbox_inches='tight')
